#include "common.h"
#include "dipsw.h"
#include "leds.h"
#include "mb.h"
#include "mbutils.h"
#include "pwm.h"

/*
Template based on the DPF project:
	http://siwawi.bauing.uni-kl.de/avr_projects/arm_projects/index_cortex.html#stm32_dpf
*/



#define USE_WATCHDOG	1			//enable/disable watchdog

extern uint32_t _isr_vectorsram_offs;
extern uint32_t _isr_vectorsflash_offs;

#define CHAN_NB		10			//number of channels
#define REG_VAL_BASE	101		//base address for value registers
#define REG_DELAY_BASE	201		//base address for fade delay registers

#define INITIAL_DELAY	10

static volatile uint8_t reg_powerSet[CHAN_NB];	//set (final) value for each channel
static volatile uint8_t reg_delay[CHAN_NB];		//transition delay for each channel in [ms]
static uint8_t reg_power[CHAN_NB];				//current (actual) value in each channel
static uint8_t delaycntr[CHAN_NB];				//software timer counters for managing transitions

void ledTask(void)
{
	for(uint32_t i=0; i<CHAN_NB; i++)
	{
		if(reg_delay[i] == 0)
		{
			reg_power[i] = reg_powerSet[i];
			pwmSet(i,reg_power[i]);
		}
		else
		{
			if(delaycntr[i] < reg_delay[i])
			{
				delaycntr[i]++;
			}
			else
			{
				delaycntr[i] = 0;
				
				if(reg_power[i] < reg_powerSet[i])
				{
					reg_power[i]++;
				}
				
				if(reg_power[i] > reg_powerSet[i])
				{
					reg_power[i]--;
				}
			}
			
			
		}
		pwmSet(i,reg_power[i]);
	}
	
}


void init(void)
{
	SystemInit();

	//config iwdg
	#if USE_WATCHDOG
		IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
		IWDG_SetPrescaler(IWDG_Prescaler_128);
		IWDG_SetReload(0x0AFF);
		IWDG_Enable();
		IWDG_ReloadCounter();
	#endif
		
	/* Enable USART1, GPIOA, GPIOD and AFIO clocks */
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_USART1 | RCC_APB2Periph_GPIOA | RCC_APB2Periph_GPIOB
	 | RCC_APB2Periph_GPIOC | RCC_APB2Periph_AFIO, ENABLE);

	#ifdef VECT_TAB_RAM
		/* vector-offset (TBLOFF) from bottom of SRAM. defined in linker script */
		
		/* Set the Vector Table base location at 0x20000000+_isr_vectorsram_offs */
		NVIC_SetVectorTable(NVIC_VectTab_RAM, (uint32_t)&_isr_vectorsram_offs);
	#else
		
		/* Set the Vector Table base location at 0x08000000+_isr_vectorsflash_offs */
		NVIC_SetVectorTable(NVIC_VectTab_FLASH, (uint32_t)&_isr_vectorsflash_offs);
	#endif /* VECT_TAB_RAM */
	
	NVIC_PriorityGroupConfig( NVIC_PriorityGroup_4 );
	
	/* Configure HCLK clock as SysTick clock source. */
	SysTick_CLKSourceConfig( SysTick_CLKSource_HCLK );
	
	ledsInit();
	
	debug_init_default();
	
	xprintf("\n\n *** The PWMLEDz Project debug console ready ***\n");
	
	#if 1
	/* Setup SysTick Timer */
	if ( timebase_config() ) {
		xprintf("timebase_config error\n");
	}
	#endif
	
	uint8_t stAddr = dipsw_state();
	xprintf("Modbus station address = %x\n",stAddr);
	
	xprintf("Initializing FreeModbus lib...\n");
	eMBErrorCode err = eMBInit( MB_RTU, stAddr, 2, 19200, MB_PAR_NONE );
	if(err == MB_ENOERR)
	{
		xprintf("Modbus init OK\n");
	}
	else
	{
		xprintf("Modbus init failed with code %d. Program stopped.\n",(int)err);
		while(1);
	}
	
	for(int i=0;i<CHAN_NB;i++)
		reg_delay[i] = INITIAL_DELAY;
	
	pwmInit();
}


//just blink the on-board LED
void blinkImAlive(void)
{
	static uint32_t toggle=0;
	static uint32_t cntr=0;
	
	const uint32_t PERIOD = 1000;	//turns
	if(cntr < (PERIOD))
		cntr++;
	else
	{
		cntr=0;
		if(toggle) toggle = 0; else toggle = 1;
		if(toggle) led1(1); else led1(0);
	}
}


int main(void)
{
	init();
	
	eMBEnable();
	
	while(1)
	{
		eMBPoll();
		ledTask();
		delay_ms(1);
		
		blinkImAlive();
		
		#if USE_WATCHDOG
			IWDG_ReloadCounter();
		#endif
	}
	
	return 0;
}



//callback from FreeModbus library
eMBErrorCode
eMBRegHoldingCB( UCHAR * pucRegBuffer, USHORT usAddress, USHORT usNRegs, eMBRegisterMode eMode )
{
    eMBErrorCode    eStatus = MB_ENOERR;
    int             iRegIndex;

    if( ( usAddress >= REG_VAL_BASE ) && ( usAddress <= (REG_VAL_BASE + CHAN_NB) ) )
    {
        iRegIndex = ( int )( usAddress - REG_VAL_BASE );
        switch ( eMode )
        {
			case MB_REG_READ:
			{
				while( usNRegs > 0 )
				{
					uint16_t temp = reg_powerSet[iRegIndex];
					*pucRegBuffer++ = ( unsigned char )( temp >> 8 );
					*pucRegBuffer++ = ( unsigned char )( temp & 0xFF );
					iRegIndex++;
					usNRegs--;
				}
				break;
			}
			
			case MB_REG_WRITE:
			{
				while( usNRegs > 0 )
				{
					uint16_t temp;
					temp = *pucRegBuffer++ << 8;
					temp |= *pucRegBuffer++;
					if(temp > 255) temp = 255;
					reg_powerSet[iRegIndex] = temp;
					
					iRegIndex++;
					usNRegs--;
				}
				break;
			}
		}
    }
    else if( ( usAddress >= REG_DELAY_BASE ) && ( usAddress <= (REG_DELAY_BASE + CHAN_NB) ) )
    {
        iRegIndex = ( int )( usAddress - REG_DELAY_BASE );
        switch ( eMode )
        {
			case MB_REG_READ:
			{
				while( usNRegs > 0 )
				{
					uint16_t temp = reg_delay[iRegIndex];
					*pucRegBuffer++ = ( unsigned char )( temp >> 8 );
					*pucRegBuffer++ = ( unsigned char )( temp & 0xFF );
					iRegIndex++;
					usNRegs--;
				}
				break;
			}
			
			case MB_REG_WRITE:
			{
				while( usNRegs > 0 )
				{
					uint16_t temp;
					temp = *pucRegBuffer++ << 8;
					temp |= *pucRegBuffer++;
					if(temp > 255) temp = 255;
					reg_delay[iRegIndex] = temp;
					
					iRegIndex++;
					usNRegs--;
				}
			}
		}
    }
	else
    {
        eStatus = MB_ENOREG;
    }
	
    return eStatus;
}

