#include "common.h"
#include "dipsw.h"
#include "leds.h"
#include "../spp.h"
#include "pwm.h"

/*
Template based on the DPF project:
	http://siwawi.bauing.uni-kl.de/avr_projects/arm_projects/index_cortex.html#stm32_dpf
*/



#define USE_WATCHDOG	1			//enable/disable watchdog

#define SPP_RQ_VAL			0
#define SPP_RQ_DELAY		1
#define SPP_RQ_DELAY_ALT	2

volatile uint8_t sppAddress;
T_sppPacket packet;

extern uint32_t _isr_vectorsram_offs;
extern uint32_t _isr_vectorsflash_offs;

#define CHANNEL_NB		10			//number of channels

#define POWER_STEP		10
#define POWER_MAX		255

typedef struct {
	uint8_t powerSet;	//programmed (final) power level
	uint8_t powerDefault;	//default power value for on/toggle cmds
	uint8_t power;		//current power level
	uint8_t delay;		//delay for channel fade
	uint8_t counter;	//fade counter
}T_channel;

static T_channel outputs[CHANNEL_NB];

void ledTask(void)
{
	for(uint32_t i=0; i<CHANNEL_NB; i++)
	{
		if(outputs[i].delay == 0)
		{
			outputs[i].power = outputs[i].powerSet;
		}
		else
		{
			if(outputs[i].counter < outputs[i].delay)
			{
				outputs[i].counter++;
			}
			else
			{
				outputs[i].counter = 0;
				
				if(outputs[i].power < outputs[i].powerSet)
				{
					outputs[i].power++;
				}
				
				if(outputs[i].power > outputs[i].powerSet)
				{
					outputs[i].power--;
				}
			}
			
		}
		
		pwmSet(i,outputs[i].power);
	}
	
}


void init(void)
{
	SystemInit();

	//config iwdg
	#if USE_WATCHDOG
		IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
		IWDG_SetPrescaler(IWDG_Prescaler_128);
		IWDG_SetReload(0x0AFF);
		IWDG_Enable();
		IWDG_ReloadCounter();
	#endif
		
	/* Enable USART1, GPIOA, GPIOD and AFIO clocks */
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_USART1 | RCC_APB2Periph_GPIOA | RCC_APB2Periph_GPIOB
	 | RCC_APB2Periph_GPIOC | RCC_APB2Periph_AFIO, ENABLE);

	#ifdef VECT_TAB_RAM
		/* vector-offset (TBLOFF) from bottom of SRAM. defined in linker script */
		
		/* Set the Vector Table base location at 0x20000000+_isr_vectorsram_offs */
		NVIC_SetVectorTable(NVIC_VectTab_RAM, (uint32_t)&_isr_vectorsram_offs);
	#else
		
		/* Set the Vector Table base location at 0x08000000+_isr_vectorsflash_offs */
		NVIC_SetVectorTable(NVIC_VectTab_FLASH, (uint32_t)&_isr_vectorsflash_offs);
	#endif /* VECT_TAB_RAM */
	
	NVIC_PriorityGroupConfig( NVIC_PriorityGroup_4 );
	
	/* Configure HCLK clock as SysTick clock source. */
	SysTick_CLKSourceConfig( SysTick_CLKSource_HCLK );
	
	ledsInit();
	
	debug_init_default();
	
	xprintf("\n\n *** The PWMLEDz Project debug console ready ***\n");
	
	#if 1
	/* Setup SysTick Timer */
	if ( timebase_config() ) {
		xprintf("timebase_config error\n");
	}
	#endif
	
	sppAddress = dipsw_state();
	xprintf("SPP device address = %x\n",sppAddress);
	
	xprintf("Initializing SPP driver...\n");
	
	sppInit();
	
	xprintf("done!\n");
	
	pwmInit();
	
	for(int i=0;i<CHANNEL_NB;i++)
	{
		outputs[i].powerSet = 0;
		outputs[i].powerDefault = 255;
		outputs[i].power = 0;
		outputs[i].delay = 10;
		outputs[i].counter = 0;
	}
}


//just blink the on-board LED
void blinkImAlive(void)
{
	static uint32_t toggle=0;
	static uint32_t cntr=0;
	
	const uint32_t PERIOD = 1000;	//turns
	if(cntr < (PERIOD))
		cntr++;
	else
	{
		cntr=0;
		if(toggle) toggle = 0; else toggle = 1;
		if(toggle) led1(1); else led1(0);
	}
}


uint8_t sppRequestRxed(T_sppPacket* pPacket, uint8_t *reg)
{
	if((pPacket->dstAddr&0x0F) == SPP_BCAST_IN_CLASS) return 0;	//reject multicast packets
	if(pPacket->dstAddr == SPP_BCAST_ALL) return 0;			//reject broadcast packets
	
	if((pPacket->payload[0]&0xF0) == SPP_ID_REQ)
	{
		*reg = pPacket->payload[0]&0x0F;
		return 1;
	}
	return 0;
}


//0 - there was no request
//1 - there was a request but no valid response could be generated
//2 - there was a request and a valid response has been txed
int processRequest(T_sppPacket* pPacket)
{
	uint8_t reg;
	if(sppRequestRxed(pPacket,&reg))
	{
		pPacket->dstAddr = pPacket->srcAddr;
		pPacket->len = CHANNEL_NB + 1;
		pPacket->payload[0] = SPP_ID_RESP | reg;
		
		switch(reg)
		{
			case 0:	//get programmed power levels
			{
				for(int i=0;i<CHANNEL_NB;i++)
				{
					pPacket->payload[i+1] = outputs[i].powerSet;
				}
				sppTx(pPacket);
				return 2;
			}
			case 1:	//get default power levels
			{
				for(int i=0;i<CHANNEL_NB;i++)
				{
					pPacket->payload[i+1] = outputs[i].powerDefault;
				}
				sppTx(pPacket);
				return 2;
			}
			case 2:	//get current power levels
			{
				for(int i=0;i<CHANNEL_NB;i++)
				{
					pPacket->payload[i+1] = outputs[i].power;
				}
				sppTx(pPacket);
				return 2;
			}
			case 3:	//get fade delays
			{
				for(int i=0;i<CHANNEL_NB;i++)
				{
					pPacket->payload[i+1] = outputs[i].delay;
				}
				sppTx(pPacket);
				return 2;
			}
		}
		
		return 1;
		
	}
	
	return 0;
}


int processGlobalCmd(T_sppPacket* pPacket)
{
	if(pPacket->len==0) return 0;
	uint8_t cmd = pPacket->payload[0];
	switch(cmd)
	{
		case SPP_ID_ALL_OFF:		//global off fast
			for(uint8_t i=0;i<CHANNEL_NB;i++)
			{
				outputs[i].power = 0;
				outputs[i].powerSet = 0;
			}
			return 2;
		
		case SPP_ID_ALL_OFF_ALT:	//slow fade
			for(uint8_t i=0;i<CHANNEL_NB;i++)
			{
				outputs[i].powerSet = 0;
			}
			return 2;
		
		case SPP_ID_ALL_ON:
			for(uint8_t i=0;i<CHANNEL_NB;i++)
			{
				outputs[i].powerSet = outputs[i].powerDefault;
				outputs[i].power = outputs[i].powerDefault;
			}
			return 2;
		
		case SPP_ID_ALL_ON_ALT:
			for(uint8_t i=0;i<CHANNEL_NB;i++)
			{
				outputs[i].powerSet = outputs[i].powerDefault;
			}
			return 2;
	}
	return 0;
}


void writeChannel(uint8_t channel, uint8_t cmd, uint8_t param)
{
	//xprintf("wr: ch=%x, cmd=%x, p=%x\n",channel,cmd,param);
	
	if(channel > (CHANNEL_NB-1)) return;
	
	switch(cmd)
	{
		case (SPP_ID_CHAN_TOGGLE):		//toggle fast
			if(outputs[channel].powerSet)
			{
				outputs[channel].powerSet = 0;	//set final power value
				outputs[channel].power = 0;		//set current power value (do it now!)
			}
			else
			{
				outputs[channel].powerSet = outputs[channel].powerDefault;
				outputs[channel].power = outputs[channel].powerDefault;
			}
			break;
		case (SPP_ID_CHAN_TOGGLE_ALT):		//toggle with fade
			if(outputs[channel].powerSet)
			{
				outputs[channel].powerSet = 0;
			}
			else
			{
				outputs[channel].powerSet = outputs[channel].powerDefault;
			}
			break;
		case (SPP_ID_CHAN_UP):		//manual dimmer up
			if( ((uint16_t)outputs[channel].powerSet + POWER_STEP) <= POWER_MAX )
				outputs[channel].powerSet += POWER_STEP;
			else
				outputs[channel].powerSet += POWER_MAX;
			break;
		case (SPP_ID_CHAN_DN):		//manual dimmer down
			if( outputs[channel].powerSet >= POWER_STEP )
				outputs[channel].powerSet -= POWER_STEP;
			else
				outputs[channel].powerSet = 0;
			break;
		case(SPP_ID_CHAN_ON):	//just on
			outputs[channel].powerSet = POWER_MAX;
			outputs[channel].power = POWER_MAX;
			break;
		case(SPP_ID_CHAN_OFF):	//just off
			outputs[channel].powerSet = 0;
			outputs[channel].power = 0;
			break;
		case(SPP_ID_CHAN_ON_ALT):	//slow on
			outputs[channel].powerSet = POWER_MAX;
			break;
		case(SPP_ID_CHAN_OFF_ALT):	//slow off
			outputs[channel].powerSet = 0;
			break;
		case(SPP_ID_CHAN_VALUE):	//arbitrary value normal
			outputs[channel].powerSet = param;
			outputs[channel].power = param;
			break;
		case(SPP_ID_CHAN_VALUE_ALT):	//arbitrary value alt
			outputs[channel].powerSet = param;
			break;
		case(SPP_ID_POWER_DEFAULT):
			outputs[channel].powerDefault = param;
			break;
		case(SPP_ID_TRANSITION):
			outputs[channel].delay = param;
			break;
	}
}

void exeMultichannel(uint8_t cmd, uint8_t param, uint16_t mask)
{
	//xprintf("multi: cmd=%x, p=%x, m=%x\n",cmd,param,mask);
	for(uint8_t i=0;i<CHANNEL_NB;i++)
	{
		if( ((uint16_t)1<<i) & mask )	//if channel is selected
		{
			writeChannel( i, cmd, param );
		}
	}
}

int processMultiChannel(T_sppPacket* pPacket)
{
	if(pPacket->payload[0]==SPP_ID_MULTICHANNEL)
	{
		if(pPacket->len < 4) return 0;	//a simple sanity check
		uint16_t chanMask = ((uint16_t)pPacket->payload[1]<<8) | pPacket->payload[2];
		uint8_t cmd = pPacket->payload[3] & 0xF0;	//ignore bits [3:0] because the channel info is in the mask
		uint8_t param = pPacket->payload[4];
		exeMultichannel(cmd,param,chanMask);
		return 1;
	}
	return 0;
}


void processChannelCommand(T_sppPacket* pPacket)
{
	if(pPacket->len == 0) return;
	
	uint8_t cmd = pPacket->payload[0] & 0xF0;
	uint8_t channel = pPacket->payload[0] & 0x0F;	//the first channel for this command
	
	switch(cmd)
	{
		case SPP_ID_CHAN_VALUE:
		case SPP_ID_CHAN_VALUE_ALT:
		case SPP_ID_POWER_DEFAULT:
		case SPP_ID_TRANSITION:
		{
			uint8_t dataLen = pPacket->len-1;
			for(uint8_t i=0;i<dataLen;i++)
			{
				uint8_t param = pPacket->payload[i+1];
				writeChannel(i+channel,cmd,param);
			}
			break;
		}
		default:
		{
			writeChannel(channel,cmd,0);
			break;
		}
	}
	
}

void processRxedPacket(T_sppPacket* pPacket)
{
	if(processRequest(pPacket) > 0) return;
	if(processMultiChannel(pPacket) > 0) return;
	if(processGlobalCmd(pPacket) > 0) return;
	processChannelCommand(pPacket);
}


int main(void)
{
	init();
	
	while(1)
	{
		ledTask();
		delay_ms(1);
		blinkImAlive();
		#if USE_WATCHDOG
			IWDG_ReloadCounter();
		#endif
		
		
		if(sppRx(&packet))
		{
			processRxedPacket(&packet);
		}
		
		sppIdleTimeout();
	}
	
	return 0;
}


